@@ -11,7 +11,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, see <http://www.gnu.org/licenses/>.
+ * License along with this library; if not, see <https://www.gnu.org/licenses/>.
  *
  * Copyright 2006,2007 Øyvind Kolås <pippin@gimp.org>
  */
@@ -21,9 +21,7 @@
 #include <glib.h>
 #include <glib-object.h>
 
-#include "gegl.h"
-#include "gegl-types-internal.h"
-#include "gegl-config.h"
+#include "gegl-buffer-config.h"
 #include "gegl-buffer.h"
 #include "gegl-buffer-private.h"
 #include "gegl-tile.h"
@@ -31,12 +29,15 @@
 #include "gegl-tile-storage.h"
 #include "gegl-debug.h"
 
-#include "gegl-buffer-cl-cache.h"
-
 /*
 #define GEGL_DEBUG_CACHE_HITS
 */
 
+#define GEGL_CACHE_TRIM_INTERVAL   100000 /* microseconds */
+#define GEGL_CACHE_TRIM_RATIO_MIN  0.01
+#define GEGL_CACHE_TRIM_RATIO_MAX  0.50
+#define GEGL_CACHE_TRIM_RATIO_RATE 2.0
+
 typedef struct CacheItem
 {
   GeglTile *tile; /* The tile */
@@ -54,38 +55,35 @@ typedef struct CacheItem
         ((CacheItem *) ((guchar *) l - G_STRUCT_OFFSET (CacheItem, link)))
 
 
-static gboolean   gegl_tile_handler_cache_equalfunc  (gconstpointer         a,
-                                                      gconstpointer         b);
-static guint      gegl_tile_handler_cache_hashfunc   (gconstpointer         key);
-static void       gegl_tile_handler_cache_dispose    (GObject              *object);
-static gboolean   gegl_tile_handler_cache_wash       (GeglTileHandlerCache *cache);
-static gpointer   gegl_tile_handler_cache_command    (GeglTileSource       *tile_store,
-                                                      GeglTileCommand       command,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z,
-                                                      gpointer              data);
-static GeglTile * gegl_tile_handler_cache_get_tile   (GeglTileHandlerCache *cache,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z);
-static gboolean   gegl_tile_handler_cache_has_tile   (GeglTileHandlerCache *cache,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z);
-void              gegl_tile_handler_cache_insert     (GeglTileHandlerCache *cache,
-                                                      GeglTile             *tile,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z);
-static void       gegl_tile_handler_cache_void       (GeglTileHandlerCache *cache,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z);
-static void       gegl_tile_handler_cache_invalidate (GeglTileHandlerCache *cache,
-                                                      gint                  x,
-                                                      gint                  y,
-                                                      gint                  z);
+static gboolean   gegl_tile_handler_cache_equalfunc  (gconstpointer             a,
+                                                      gconstpointer             b);
+static guint      gegl_tile_handler_cache_hashfunc   (gconstpointer             key);
+static void       gegl_tile_handler_cache_dispose    (GObject                  *object);
+static gboolean   gegl_tile_handler_cache_wash       (GeglTileHandlerCache     *cache);
+static gpointer   gegl_tile_handler_cache_command    (GeglTileSource           *tile_store,
+                                                      GeglTileCommand           command,
+                                                      gint                      x,
+                                                      gint                      y,
+                                                      gint                      z,
+                                                      gpointer                  data);
+static gboolean   gegl_tile_handler_cache_has_tile   (GeglTileHandlerCache     *cache,
+                                                      gint                      x,
+                                                      gint                      y,
+                                                      gint                      z);
+static void       gegl_tile_handler_cache_void       (GeglTileHandlerCache     *cache,
+                                                      gint                      x,
+                                                      gint                      y,
+                                                      gint                      z,
+                                                      guint64                   damage);
+static void       gegl_tile_handler_cache_invalidate (GeglTileHandlerCache     *cache,
+                                                      gint                      x,
+                                                      gint                      y,
+                                                      gint                      z);
+static gboolean   gegl_tile_handler_cache_copy       (GeglTileHandlerCache     *cache,
+                                                      gint                      x,
+                                                      gint                      y,
+                                                      gint                      z,
+                                                      const GeglTileCopyParams *params);
 
 
 static GMutex             mutex                 = { 0, };
@@ -116,11 +114,8 @@ gegl_tile_handler_cache_init (GeglTileHandlerCache *cache)
   ((GeglTileSource*)cache)->command = gegl_tile_handler_cache_command;
   cache->items = g_hash_table_new (gegl_tile_handler_cache_hashfunc, gegl_tile_handler_cache_equalfunc);
   g_queue_init (&cache->queue);
-  gegl_tile_cache_init ();
 
-  g_mutex_lock (&mutex);
-  g_queue_push_tail_link (&cache_queue, &cache->link);
-  g_mutex_unlock (&mutex);
+  gegl_tile_handler_cache_connect (cache);
 }
 
 static void
@@ -130,21 +125,10 @@ drop_hot_tile (GeglTile *tile)
 
   if (storage)
     {
-#if 0 /* the storage's mutex should already be locked at this point */
-      if (gegl_config_threads()>1)
-        g_rec_mutex_lock (&storage->mutex);
-#endif
+      tile = gegl_tile_storage_try_steal_hot_tile (storage, tile);
 
-      if (storage->hot_tile == tile)
-        {
-          gegl_tile_unref (storage->hot_tile);
-          storage->hot_tile = NULL;
-        }
-
-#if 0
-      if (gegl_config_threads()>1)
-        g_rec_mutex_unlock (&storage->mutex);
-#endif
+      if (tile)
+        gegl_tile_unref (tile);
     }
 }
 
@@ -186,9 +170,7 @@ gegl_tile_handler_cache_dispose (GObject *object)
 {
   GeglTileHandlerCache *cache = GEGL_TILE_HANDLER_CACHE (object);
 
-  g_mutex_lock (&mutex);
-  g_queue_unlink (&cache_queue, &cache->link);
-  g_mutex_unlock (&mutex);
+  gegl_tile_handler_cache_disconnect (cache);
 
   gegl_tile_handler_cache_reinit (cache);
 
@@ -206,8 +188,8 @@ gegl_tile_handler_cache_get_tile_command (GeglTileSource *tile_store,
   GeglTileSource       *source   = ((GeglTileHandler*) (tile_store))->source;
   GeglTile             *tile     = NULL;
 
-  if (G_UNLIKELY (gegl_cl_is_accelerated ()))
-    gegl_buffer_cl_cache_flush2 (cache, NULL);
+  if (gegl_tile_handler_cache_ext_flush)
+    gegl_tile_handler_cache_ext_flush (cache, NULL);
 
   tile = gegl_tile_handler_cache_get_tile (cache, x, y, z);
   if (tile)
@@ -246,8 +228,8 @@ gegl_tile_handler_cache_command (GeglTileSource  *tile_store,
         {
           GList *link;
 
-          if (gegl_cl_is_accelerated ())
-            gegl_buffer_cl_cache_flush2 (cache, NULL);
+          if (gegl_tile_handler_cache_ext_flush)
+            gegl_tile_handler_cache_ext_flush (cache, NULL);
 
           for (link = g_queue_peek_head_link (&cache->queue);
                link;
@@ -287,11 +269,19 @@ gegl_tile_handler_cache_command (GeglTileSource  *tile_store,
         gegl_tile_handler_cache_invalidate (cache, x, y, z);
         break;
       case GEGL_TILE_VOID:
-        gegl_tile_handler_cache_void (cache, x, y, z);
+        gegl_tile_handler_cache_void (cache, x, y, z,
+                                      data ? *(const guint64 *) data :
+                                             ~(guint64) 0);
         break;
       case GEGL_TILE_REINIT:
         gegl_tile_handler_cache_reinit (cache);
         break;
+      case GEGL_TILE_COPY:
+        /* we potentially chain up in gegl_tile_handler_cache_copy(), because
+         * its logic is interleaved with that of the backend.
+         */
+        return GINT_TO_POINTER (gegl_tile_handler_cache_copy (cache,
+                                                              x, y, z, data));
       default:
         break;
     }
@@ -333,6 +323,10 @@ gegl_tile_handler_cache_find_oldest_cache (GeglTileHandlerCache *prev_cache)
       if (! time)
         continue;
 
+      /* the cache is being disconnected */
+      if (! cache->link.data)
+        continue;
+
       if (time == stamp)
         {
           oldest_cache = cache;
@@ -412,8 +406,7 @@ gegl_tile_handler_cache_wash (GeglTileHandlerCache *cache)
       if (cache == NULL)
         break;
 
-      if (gegl_config_threads()>1 &&
-          ! g_rec_mutex_trylock (&cache->tile_storage->mutex))
+      if (! g_rec_mutex_trylock (&cache->tile_storage->mutex))
         {
           continue;
         }
@@ -438,8 +431,7 @@ gegl_tile_handler_cache_wash (GeglTileHandlerCache *cache)
           size += tile->size;
         }
 
-      if (gegl_config_threads()>1)
-        g_rec_mutex_unlock (&cache->tile_storage->mutex);
+      g_rec_mutex_unlock (&cache->tile_storage->mutex);
     }
 
   g_mutex_unlock (&mutex);
@@ -471,7 +463,7 @@ cache_lookup (GeglTileHandlerCache *cache,
 
 /* returns the requested Tile if it is in the cache, NULL otherwize.
  */
-static GeglTile *
+GeglTile *
 gegl_tile_handler_cache_get_tile (GeglTileHandlerCache *cache,
                                   gint                  x,
                                   gint                  y,
@@ -520,31 +512,61 @@ gegl_tile_handler_cache_has_tile (GeglTileHandlerCache *cache,
 static gboolean
 gegl_tile_handler_cache_trim (GeglTileHandlerCache *cache)
 {
-
-  GList *link;
+  GList          *link;
+  gint64          time;
+  static gint64   last_time;
+  static gdouble  ratio  = GEGL_CACHE_TRIM_RATIO_MIN;
+  guint64         target_size;
+  static guint    counter;
 
   cache = NULL;
   link  = NULL;
 
   g_mutex_lock (&mutex);
 
-  while ((guintptr) g_atomic_pointer_get (&cache_total) >
-         gegl_config ()->tile_cache_size)
+  target_size = gegl_buffer_config ()->tile_cache_size;
+
+  if ((guintptr) g_atomic_pointer_get (&cache_total) <= target_size)
+    {
+      g_mutex_unlock (&mutex);
+
+      return TRUE;
+    }
+
+  time = g_get_monotonic_time ();
+
+  if (time - last_time < GEGL_CACHE_TRIM_INTERVAL)
+    {
+      ratio = MIN (ratio * GEGL_CACHE_TRIM_RATIO_RATE,
+                   GEGL_CACHE_TRIM_RATIO_MAX);
+    }
+  else if (time - last_time >= 2 * GEGL_CACHE_TRIM_INTERVAL)
+    {
+      ratio = GEGL_CACHE_TRIM_RATIO_MIN;
+    }
+
+  target_size -= target_size * ratio;
+
+  g_mutex_unlock (&mutex);
+
+  while ((guintptr) g_atomic_pointer_get (&cache_total) > target_size)
     {
       CacheItem *last_writable;
       GeglTile  *tile;
       GList     *prev_link;
 
 #ifdef GEGL_DEBUG_CACHE_HITS
-      GEGL_NOTE(GEGL_DEBUG_CACHE, "cache_total:"G_GUINT64_FORMAT" > cache_size:"G_GUINT64_FORMAT, cache_total, gegl_config()->tile_cache_size);
+      GEGL_NOTE(GEGL_DEBUG_CACHE, "cache_total:"G_GUINT64_FORMAT" > cache_size:"G_GUINT64_FORMAT, cache_total, gegl_buffer_config()->tile_cache_size);
       GEGL_NOTE(GEGL_DEBUG_CACHE, "%f%% hit:%i miss:%i  %i]", cache_hits*100.0/(cache_hits+cache_misses), cache_hits, cache_misses, g_queue_get_length (&cache_queue));
 #endif
 
       if (! link)
         {
-          if (cache && gegl_config_threads()>1)
+          if (cache)
             g_rec_mutex_unlock (&cache->tile_storage->mutex);
 
+          g_mutex_lock (&mutex);
+
           do
             {
               cache = gegl_tile_handler_cache_find_oldest_cache (cache);
@@ -558,15 +580,12 @@ gegl_tile_handler_cache_trim (GeglTileHandlerCache *cache)
                   * thread.  try locking the cache's storage mutex here, and
                   * skip the cache if it fails.
                   */
-                 gegl_config_threads()>1 &&
                  ! g_rec_mutex_trylock (&cache->tile_storage->mutex));
 
-          if (! cache)
-            {
-              g_mutex_unlock (&mutex);
+          g_mutex_unlock (&mutex);
 
-              return FALSE;
-            }
+          if (! cache)
+            break;
 
           link = g_queue_peek_tail_link (&cache->queue);
         }
@@ -584,9 +603,35 @@ gegl_tile_handler_cache_trim (GeglTileHandlerCache *cache)
           if (tile->ref_count > 1)
             continue;
 
+          /* if we need to maintain the tile's data-pointer identity we can't
+           * remove it from the cache, since the storage might copy the data
+           * and throw the tile away.
+           */
+          if (tile->keep_identity)
+            continue;
+
+          /* a set of cloned tiles is only counted once toward the total cache
+           * size, so the entire set has to be removed from the cache in order
+           * to reclaim the memory of a single tile.  in other words, in a set
+           * of n cloned tiles, we can assume that each individual tile
+           * contributes only 1/n of its size to the total cache size.  on the
+           * other hand, storing a cloned tile is as expensive as storing an
+           * uncloned tile.  therefore, if the tile needs to be stored, we only
+           * remove it with a probability of 1/n.
+           */
+          if (gegl_tile_needs_store (tile) &&
+              counter++ % *gegl_tile_n_cached_clones (tile))
+            {
+              continue;
+            }
+
           break;
         }
 
+      /* the cache is being disconnected */
+      if (! cache->link.data)
+        link = NULL;
+
       if (! link)
         continue;
 
@@ -611,12 +656,16 @@ gegl_tile_handler_cache_trim (GeglTileHandlerCache *cache)
       link = prev_link;
     }
 
-  if (cache && gegl_config_threads()>1)
+  if (cache)
     g_rec_mutex_unlock (&cache->tile_storage->mutex);
 
+  g_mutex_lock (&mutex);
+
+  last_time = g_get_monotonic_time ();
+
   g_mutex_unlock (&mutex);
 
-  return TRUE;
+  return cache != NULL;
 }
 
 static void
@@ -649,34 +698,170 @@ gegl_tile_handler_cache_invalidate (GeglTileHandlerCache *cache,
     }
 }
 
+static gboolean
+gegl_tile_handler_cache_copy (GeglTileHandlerCache     *cache,
+                              gint                      x,
+                              gint                      y,
+                              gint                      z,
+                              const GeglTileCopyParams *params)
+{
+  GeglTile *tile;
+  GeglTile *dst_tile = NULL;
+  gboolean  success  = FALSE;
+
+  if (gegl_tile_handler_cache_ext_flush)
+    gegl_tile_handler_cache_ext_flush (cache, NULL);
+
+  tile = gegl_tile_handler_cache_get_tile (cache, x, y, z);
+
+  /* if the tile is not fully valid, bail, so that the copy happens using a
+   * TILE_GET commands, validating the tile in the process.
+   */
+  if (tile && tile->damage)
+    {
+      gegl_tile_unref (tile);
+
+      return FALSE;
+    }
+
+  /* otherwise, if we have the tile and it's valid, copy it directly to the
+   * destination first, so that the copied tile will already be present in the
+   * destination cache.
+   */
+  if (tile)
+    {
+      GeglTileHandler *dst_handler;
+
+      if (params->dst_buffer)
+        dst_handler = GEGL_TILE_HANDLER (params->dst_buffer->tile_storage);
+      else
+        dst_handler = GEGL_TILE_HANDLER (cache);
+
+      dst_tile = gegl_tile_handler_dup_tile (dst_handler,
+                                             tile,
+                                             params->dst_x,
+                                             params->dst_y,
+                                             params->dst_z);
+
+      success = TRUE;
+    }
+
+  /* then, if we don't have the tile, or if the tile is stored, iow, if the
+   * tile might exist in the backend in an up-to-date state ...
+   */
+  if (! tile || gegl_tile_is_stored (tile))
+    {
+      /* ... try letting the backend copy the tile, so that the copied tile
+       * will already be stored in the destination backend.
+       */
+      if (gegl_tile_handler_source_command (GEGL_TILE_HANDLER (cache),
+                                            GEGL_TILE_COPY, x, y, z,
+                                            (gpointer) params))
+        {
+          /* if the backend copied the tile, we can either mark the copied tile
+           * as stored, or, if we didn't have the tile, void any existing tile
+           * in the destination cache.
+           */
+
+          if (dst_tile)
+            {
+              gegl_tile_mark_as_stored (dst_tile);
+            }
+          else
+            {
+              GeglTileHandlerCache *dst_cache;
+
+              if (params->dst_buffer)
+                dst_cache = params->dst_buffer->tile_storage->cache;
+              else
+                dst_cache = cache;
+
+              if (dst_cache)
+                {
+                  gegl_tile_handler_cache_void (dst_cache,
+                                                params->dst_x,
+                                                params->dst_y,
+                                                params->dst_z,
+                                                ~(guint64) 0);
+                }
+            }
+
+          success = TRUE;
+        }
+    }
+
+  if (dst_tile)
+    gegl_tile_unref (dst_tile);
+
+  if (tile)
+    gegl_tile_unref (tile);
+
+  /* the copy is considered successful if either we or the backend (or both)
+   * copied the tile.
+   */
+  return success;
+}
+
+
+static void
+gegl_tile_handler_cache_remove_item (GeglTileHandlerCache *cache,
+                                     CacheItem            *item)
+{
+  if (g_atomic_int_dec_and_test (gegl_tile_n_cached_clones (item->tile)))
+    g_atomic_pointer_add (&cache_total, -item->tile->size);
+  g_atomic_pointer_add (&cache_total_uncloned, -item->tile->size);
+
+  g_queue_unlink (&cache->queue, &item->link);
+  g_hash_table_remove (cache->items, item);
+
+  if (g_queue_is_empty (&cache->queue))
+    cache->time = cache->stamp = 0;
+
+  item->tile->tile_storage = NULL;
+  gegl_tile_unref (item->tile);
+
+  g_slice_free (CacheItem, item);
+}
+
+void
+gegl_tile_handler_cache_remove (GeglTileHandlerCache *cache,
+                                gint                  x,
+                                gint                  y,
+                                gint                  z)
+{
+  CacheItem *item;
+
+  item = cache_lookup (cache, x, y, z);
+
+  if (item)
+    {
+      drop_hot_tile (item->tile);
+
+      gegl_tile_handler_cache_remove_item (cache, item);
+    }
+}
 
 static void
 gegl_tile_handler_cache_void (GeglTileHandlerCache *cache,
                               gint                  x,
                               gint                  y,
-                              gint                  z)
+                              gint                  z,
+                              guint64               damage)
 {
   CacheItem *item;
 
   item = cache_lookup (cache, x, y, z);
   if (item)
     {
-      if (g_atomic_int_dec_and_test (gegl_tile_n_cached_clones (item->tile)))
-        g_atomic_pointer_add (&cache_total, -item->tile->size);
-      g_atomic_pointer_add (&cache_total_uncloned, -item->tile->size);
-
-      g_queue_unlink (&cache->queue, &item->link);
-      g_hash_table_remove (cache->items, item);
-
-      if (g_queue_is_empty (&cache->queue))
-        cache->time = cache->stamp = 0;
-
       drop_hot_tile (item->tile);
-      gegl_tile_void (item->tile);
-      item->tile->tile_storage = NULL;
-      gegl_tile_unref (item->tile);
 
-      g_slice_free (CacheItem, item);
+      if (gegl_tile_damage (item->tile, damage))
+        gegl_tile_handler_cache_remove_item (cache, item);
+    }
+  else if (z == 0 && damage)
+    {
+      gegl_tile_handler_damage_tile (GEGL_TILE_HANDLER (cache),
+                                     x, y, z, damage);
     }
 }
 
@@ -688,6 +873,7 @@ gegl_tile_handler_cache_insert (GeglTileHandlerCache *cache,
                                 gint                  z)
 {
   CacheItem *item = g_slice_new (CacheItem);
+  guintptr   total;
 
   item->tile      = gegl_tile_ref (tile);
   item->link.data = item;
@@ -698,7 +884,7 @@ gegl_tile_handler_cache_insert (GeglTileHandlerCache *cache,
   item->z         = z;
 
   // XXX : remove entry if it already exists
-  gegl_tile_handler_cache_void (cache, x, y, z);
+  gegl_tile_handler_cache_remove (cache, x, y, z);
 
   tile->x = x;
   tile->y = y;
@@ -710,12 +896,15 @@ gegl_tile_handler_cache_insert (GeglTileHandlerCache *cache,
   cache->time = ++cache_time;
 
   if (g_atomic_int_add (gegl_tile_n_cached_clones (tile), 1) == 0)
-    g_atomic_pointer_add (&cache_total, tile->size);
+    total = g_atomic_pointer_add (&cache_total, tile->size) + tile->size;
+  else
+    total = (guintptr) g_atomic_pointer_get (&cache_total);
   g_atomic_pointer_add (&cache_total_uncloned, tile->size);
   g_hash_table_insert (cache->items, item, item);
   g_queue_push_head_link (&cache->queue, &item->link);
 
-  gegl_tile_handler_cache_trim (cache);
+  if (total > gegl_buffer_config ()->tile_cache_size)
+    gegl_tile_handler_cache_trim (cache);
 
   /* there's a race between this assignment, and the one at the bottom of
    * gegl_tile_handler_cache_tile_uncloned().  this is acceptable, though,
@@ -734,7 +923,7 @@ gegl_tile_handler_cache_tile_uncloned (GeglTileHandlerCache *cache,
   total = (guintptr) g_atomic_pointer_add (&cache_total, tile->size) +
           tile->size;
 
-  if (total > gegl_config ()->tile_cache_size)
+  if (total > gegl_buffer_config ()->tile_cache_size)
     gegl_tile_handler_cache_trim (cache);
 
   cache_total_max = MAX (cache_total_max, total);
@@ -746,6 +935,38 @@ gegl_tile_handler_cache_new (void)
   return g_object_new (GEGL_TYPE_TILE_HANDLER_CACHE, NULL);
 }
 
+void
+gegl_tile_handler_cache_connect (GeglTileHandlerCache *cache)
+{
+  /* join the global cache queue */
+  if (! cache->link.data)
+    {
+      cache->link.data = cache;
+
+      g_mutex_lock (&mutex);
+      g_queue_push_tail_link (&cache_queue, &cache->link);
+      g_mutex_unlock (&mutex);
+    }
+}
+
+void
+gegl_tile_handler_cache_disconnect (GeglTileHandlerCache *cache)
+{
+  /* leave the global cache queue */
+  if (cache->link.data)
+    {
+      cache->link.data = NULL;
+
+      g_rec_mutex_lock (&cache->tile_storage->mutex);
+
+      g_mutex_lock (&mutex);
+      g_queue_unlink (&cache_queue, &cache->link);
+      g_mutex_unlock (&mutex);
+
+      g_rec_mutex_unlock (&cache->tile_storage->mutex);
+    }
+}
+
 gsize
 gegl_tile_handler_cache_get_total (void)
 {
@@ -759,7 +980,7 @@ gegl_tile_handler_cache_get_total_max (void)
 }
 
 gsize
-gegl_tile_handler_cache_get_total_uncloned (void)
+gegl_tile_handler_cache_get_total_uncompressed (void)
 {
   return cache_total_uncloned;
 }
@@ -825,14 +1046,32 @@ gegl_tile_handler_cache_equalfunc (gconstpointer a,
   return FALSE;
 }
 
+static void
+gegl_buffer_config_tile_cache_size_notify (GObject    *gobject,
+                                           GParamSpec *pspec,
+                                           gpointer    user_data)
+{
+  if ((guintptr) g_atomic_pointer_get (&cache_total) >
+      gegl_buffer_config () ->tile_cache_size)
+    {
+      gegl_tile_handler_cache_trim (NULL);
+    }
+}
+
 void
 gegl_tile_cache_init (void)
 {
+  g_signal_connect (gegl_buffer_config (), "notify::tile-cache-size",
+                    G_CALLBACK (gegl_buffer_config_tile_cache_size_notify), NULL);
 }
 
 void
 gegl_tile_cache_destroy (void)
 {
+  g_signal_handlers_disconnect_by_func (gegl_buffer_config(),
+                                        gegl_buffer_config_tile_cache_size_notify,
+                                        NULL);
+
   g_warn_if_fail (g_queue_is_empty (&cache_queue));
   g_queue_clear (&cache_queue);
 }
