@@ -11,7 +11,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with GEGL; if not, see <http://www.gnu.org/licenses/>.
+ * License along with GEGL; if not, see <https://www.gnu.org/licenses/>.
  *
  * Copyright 2003 Calvin Williamson
  *           2006 Øyvind Kolås
@@ -35,13 +35,14 @@
 #include "gegl-visitable.h"
 #include "gegl-config.h"
 
-#include "buffer/gegl-region.h"
+#include "gegl-region.h"
 
 #include "graph/gegl-visitor.h"
 #include "graph/gegl-callback-visitor.h"
 #include "graph/gegl-node-output-visitable.h"
 
 #include "operation/gegl-operation.h"
+#include "operation/gegl-operation-private.h"
 #include "operation/gegl-operations.h"
 #include "operation/gegl-operation-meta.h"
 
@@ -54,6 +55,7 @@ enum
   PROP_OPERATION,
   PROP_NAME,
   PROP_DONT_CACHE,
+  PROP_CACHE_POLICY,
   PROP_USE_OPENCL,
   PROP_PASSTHROUGH
 };
@@ -121,6 +123,7 @@ static void            gegl_node_update_debug_name        (GeglNode *node);
 
 
 G_DEFINE_TYPE_WITH_CODE (GeglNode, gegl_node, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (GeglNode)
                          G_IMPLEMENT_INTERFACE (GEGL_TYPE_VISITABLE,
                                                 gegl_node_visitable_iface_init))
 
@@ -130,8 +133,6 @@ gegl_node_class_init (GeglNodeClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
-  g_type_class_add_private (klass, sizeof (GeglNodePrivate));
-
   gobject_class->finalize     = gegl_node_finalize;
   gobject_class->dispose      = gegl_node_dispose;
   gobject_class->set_property = gegl_node_local_set_property;
@@ -156,11 +157,21 @@ gegl_node_class_init (GeglNodeClass *klass)
   g_object_class_install_property (gobject_class, PROP_DONT_CACHE,
                                    g_param_spec_boolean ("dont-cache",
                                                          "Do not cache",
-                                                        "Do not cache the result of this operation, the property is inherited by children created from a node.",
+                                                        "Do not cache the result of this operation, the property is inherited by children created from a node."
+                                                        " (Deprecated for \"cache-policy\".)",
                                                         FALSE,
                                                         G_PARAM_CONSTRUCT |
                                                         G_PARAM_READWRITE));
 
+  g_object_class_install_property (gobject_class, PROP_CACHE_POLICY,
+                                   g_param_spec_enum ("cache-policy",
+                                                      "Cache Policy",
+                                                      "Cache policy for this node, the property is inherited by children created from a node.",
+                                                      GEGL_TYPE_CACHE_POLICY,
+                                                      GEGL_CACHE_POLICY_AUTO,
+                                                      G_PARAM_CONSTRUCT |
+                                                      G_PARAM_READWRITE));
+
   g_object_class_install_property (gobject_class, PROP_USE_OPENCL,
                                    g_param_spec_boolean ("use-opencl",
                                                          "Use OpenCL",
@@ -219,9 +230,7 @@ gegl_node_class_init (GeglNodeClass *klass)
 static void
 gegl_node_init (GeglNode *self)
 {
-  self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
-                                            GEGL_TYPE_NODE,
-                                            GeglNodePrivate);
+  self->priv = gegl_node_get_instance_private (self);
 
   self->pads             = NULL;
   self->input_pads       = NULL;
@@ -303,6 +312,10 @@ gegl_node_local_set_property (GObject      *gobject,
         node->dont_cache = g_value_get_boolean (value);
         break;
 
+      case PROP_CACHE_POLICY:
+        node->cache_policy = g_value_get_enum (value);
+        break;
+
       case PROP_PASSTHROUGH:
         node->passthrough = g_value_get_boolean (value);
         break;
@@ -353,6 +366,10 @@ gegl_node_local_get_property (GObject    *gobject,
         g_value_set_boolean (value, node->dont_cache);
         break;
 
+      case PROP_CACHE_POLICY:
+        g_value_set_enum (value, node->cache_policy);
+        break;
+
       case PROP_PASSTHROUGH:
         g_value_set_boolean (value, node->passthrough);
         break;
@@ -831,6 +848,14 @@ gegl_node_has_source (GeglNode *self,
   return found;
 }
 
+gboolean
+gegl_node_is_graph (GeglNode *node)
+{
+  g_return_val_if_fail (node, FALSE);
+  g_return_val_if_fail (GEGL_IS_NODE (node), FALSE);
+  return node->is_graph;
+}
+
 gboolean
 gegl_node_connect_from (GeglNode    *sink,
                         const gchar *sink_pad_name,
@@ -1113,7 +1138,7 @@ gegl_node_blit_buffer (GeglNode            *self,
 
   if (result)
     {
-      if (buffer)
+      if (buffer && buffer != result)
         gegl_buffer_copy (result, &request, GEGL_ABYSS_NONE, buffer, NULL);
       g_object_unref (result);
     }
@@ -1121,10 +1146,7 @@ gegl_node_blit_buffer (GeglNode            *self,
 
 static inline gboolean gegl_mipmap_rendering_enabled (void)
 {
-  static int enabled = -1;
-  if (enabled == -1)
-    enabled = g_getenv("GEGL_MIPMAP_RENDERING")!=NULL;
-  return enabled;
+  return gegl_config()->mipmap_rendering;
 }
 
 void
@@ -1136,19 +1158,22 @@ gegl_node_blit (GeglNode            *self,
                 gint                 rowstride,
                 GeglBlitFlags        flags)
 {
+  gint interpolation = flags & (GEGL_BUFFER_FILTER_ALL);
+  flags &= 0xf;
+
   g_return_if_fail (GEGL_IS_NODE (self));
   g_return_if_fail (roi != NULL);
 
   if (rowstride == GEGL_AUTO_ROWSTRIDE && format)
     rowstride = babl_format_get_bytes_per_pixel (format) * roi->width;
 
-  if (!flags)
+  if (flags == 0) // DEFAULT, just render, caching only if graph is explicitly caching itself
     {
       GeglBuffer *buffer;
 
       if (scale != 1.0)
         {
-          const GeglRectangle unscaled_roi = _gegl_get_required_for_scale (format, roi, scale);
+          const GeglRectangle unscaled_roi = _gegl_get_required_for_scale (roi, scale);
 
           buffer = gegl_node_apply_roi (self, &unscaled_roi,
               gegl_mipmap_rendering_enabled()?gegl_level_from_scale (scale):0);
@@ -1158,7 +1183,7 @@ gegl_node_blit (GeglNode            *self,
           buffer = gegl_node_apply_roi (self, roi, 0);
         }
       if (buffer && destination_buf)
-        gegl_buffer_get (buffer, roi, scale, format, destination_buf, rowstride, GEGL_ABYSS_NONE);
+        gegl_buffer_get (buffer, roi, scale, format, destination_buf, rowstride, GEGL_ABYSS_NONE | interpolation);
 
       g_clear_object (&buffer);
     }
@@ -1177,7 +1202,7 @@ gegl_node_blit (GeglNode            *self,
         {
           if (scale != 1.0)
             {
-              const GeglRectangle unscaled_roi = _gegl_get_required_for_scale (format, roi, scale);
+              const GeglRectangle unscaled_roi = _gegl_get_required_for_scale (roi, scale);
               gint  level = gegl_mipmap_rendering_enabled()?gegl_level_from_scale (scale):0;
 
               gegl_node_blit_buffer (self, buffer, &unscaled_roi, level, GEGL_ABYSS_NONE);
@@ -1194,7 +1219,7 @@ gegl_node_blit (GeglNode            *self,
         {
           gegl_buffer_get (buffer, roi, scale,
                            format, destination_buf, rowstride,
-                           GEGL_ABYSS_NONE);
+                           GEGL_ABYSS_NONE|interpolation);
         }
     }
 }
@@ -1970,6 +1995,31 @@ gegl_node_emit_computed (GeglNode *node,
   g_signal_emit (node, gegl_node_signals[COMPUTED], 0, rect, NULL, NULL);
 }
 
+gboolean
+gegl_node_use_cache (GeglNode *node)
+{
+  g_return_val_if_fail (GEGL_IS_NODE (node), FALSE);
+
+  switch (node->cache_policy)
+    {
+    case GEGL_CACHE_POLICY_AUTO:
+      if (node->dont_cache)
+        return FALSE;
+      else if (node->operation)
+        return gegl_operation_use_cache (node->operation);
+      else
+        return FALSE;
+
+    case GEGL_CACHE_POLICY_NEVER:
+      return FALSE;
+
+    case GEGL_CACHE_POLICY_ALWAYS:
+      return TRUE;
+    }
+
+  g_return_val_if_reached (FALSE);
+}
+
 GeglCache *
 gegl_node_get_cache (GeglNode *node)
 {
@@ -2084,8 +2134,9 @@ gegl_node_add_child (GeglNode *self,
   self->is_graph      = TRUE;
   child->priv->parent = self;
 
-  child->dont_cache = self->dont_cache;
-  child->use_opencl = self->use_opencl;
+  child->dont_cache   = self->dont_cache;
+  child->cache_policy = self->cache_policy;
+  child->use_opencl   = self->use_opencl;
 
   return child;
 }
@@ -2194,8 +2245,9 @@ gegl_node_create_child (GeglNode    *self,
   ret = gegl_node_new_child (self, "operation", operation, NULL);
   if (ret && self)
     {
-      ret->dont_cache = self->dont_cache;
-      ret->use_opencl = self->use_opencl;
+      ret->dont_cache   = self->dont_cache;
+      ret->cache_policy = self->cache_policy;
+      ret->use_opencl   = self->use_opencl;
     }
   return ret;
 }
